/*
 * $Id: adffs.c  Handles UAE .ADF,.ADZ,.DMS,.HDF and .DMP file support within Midnight Commander Exp $
 *
 * 18/01/2003 22:59 pbakota <bakota@tippnet.co.yu>
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <errno.h>

#define VERSION "1.0a"

#include "adf_lib/Lib/adflib.h"

#ifdef DEBUG
#include "debug.h"
#define D(x) x
#else
#define D(x)
#endif

#ifdef PACKER
#include "support.h"
#endif

/* The portable way used to create a directory is to call the MKDIR command via the
 * system() function.
 * It is used to create the 'dir1' directory, like the 'dir1/dir11' directory
 */

/* the portable way to check if a directory 'dir1' already exists i'm using is to
 * do fopen('dir1','rb'). NULL is returned if 'dir1' doesn't exists yet, an handle instead
 */
#define MKDIR "mkdir"

#ifdef WIN32
#define DIRSEP '\\'
#else
#define DIRSEP '/'
#endif /* WIN32 */

#define EXTBUFL 1024*8

static char *progname;

#ifdef DEBUG

#endif

void usage(void)
{
	fprintf(stderr, "\nAdffs v" VERSION " (c) 2003 Peter Bakota <bakota@tippnet.co.yu>\n");
	fprintf(stderr, "Midnight Commander support for .ADF, .ADZ, .HDF, .HDZ, .DMP and .DMS files\n");
	fprintf(stderr, "Powered by:\n");
	fprintf(stderr, "   ADFlib (v%s - %s) (C) 1997-1998 Laurent Clevy\n", adfGetVersionNumber(), adfGetVersionDate());
	fprintf(stderr, "   xDMS by Andre Rodrigues de la Rocha <adlroc@usa.net>\n\n");
	fprintf(stderr, "Usage: %s command dumpname.adf [additional-parameters]\n", progname);
	fprintf(stderr, "Commands:\n");
	fprintf(stderr, "  list, view                       - Show directory tree\n");
	fprintf(stderr, "  copyout <amiga_file> <dest_file> - Extract file\n\n");
	exit(0);
}

void printEnt(struct Volume *vol, struct Entry* entry, char *path)
{
	/* do not print the links entries, ADFlib do not support them yet properly */
	if (entry->type==ST_LFILE || entry->type==ST_LDIR || entry->type==ST_LSOFT)
		return;

	if (entry->type==ST_DIR)
		printf("drwx");
	else
		printf("-rw-");

	printf("r--r--   1      500      500 %8ld ", entry->size);

	printf("%02d-%02d-%04d %02d:%02d ", entry->month, entry->days, entry->year,
		entry->hour, entry->mins);

	if (strlen(path)>0)
		printf("%s/", path);
	if (entry->type==ST_DIR)
		printf("%s/", entry->name);
	else
		printf("%s",entry->name);
	putchar('\n');
}

void printTree(struct Volume *vol, struct List* tree, char* path)
{
    char *buf;
    struct Entry* entry;

    while(tree) {
        printEnt(vol, tree->content, path);
        if (tree->subdir!=NULL) {
            entry = (struct Entry*)tree->content;
            if (strlen(path)>0) {
                buf=(char*)malloc(sizeof(char)* (strlen(path)+1+strlen(entry->name)+1) );
                if (!buf) {
                    fprintf(stderr,"printTree : malloc error\n");
                    return;
                }
                sprintf(buf,"%s/%s", path, entry->name);
                printTree(vol, tree->subdir, buf);
                free(buf);
            }
            else
                printTree(vol, tree->subdir, entry->name);
        }
        tree = tree->next;
    }
}

void adffs_list(struct Volume *vol)
{
	BOOL true = TRUE;
	struct List *list, *cell;

	if (isDIRCACHE(vol->dosType)) {
		adfChgEnvProp(PR_USEDIRC,&true);
	}
	cell = list = adfGetRDirEnt(vol,vol->curDirPtr,TRUE);
	printTree(vol, cell, "");
	adfFreeDirList(list);
}

BOOL isDirExists(char *name)
{
	BOOL isDir = FALSE;
	struct stat *buf = (struct stat *)malloc(sizeof(struct stat));
	if(buf) {
		if(stat(name, buf)) {
			isDir = S_ISDIR(buf->st_mode);
		}
		free(buf);
	}
	return isDir;
}

void adffs_copyout(struct Volume *vol, char *sname, char *dname, unsigned char *extbuf)
{
	char *p, name[160];
	int n;
	FILE *out;
	struct File *file;

#ifdef DEBUG
	log("%s, %s\n", sname, dname);
#endif

	// Searching for amiga dir
	p = sname;
	n = 0;
	while(*p!=0) {
		if(*p!=DIRSEP)
			name[n++]=*p;
		else {
			name[n] = '\0';
			n = 0;
			if(adfChangeDir(vol, name)!=RC_OK) return;
		}
		++p;
	}
	name[n] = '\0';
	file = adfOpenFile(vol, name, "r");
	if(!file)
		return;

	out = fopen(dname, "wb");
	if(!out) {
		BOOL ok = FALSE;
		if(errno == ENOENT) {
			char *dirname = strdup(dname);
			p = strrchr(dirname, DIRSEP); *p = '\0';
			if(!isDirExists(dirname)) {
				mkdir(dirname, 0744);
				ok = TRUE;
			}
			free(dirname);
		}
		if((out = fopen(dname, "wb"))!=NULL) ok = TRUE;
		if(!ok) return;
	}

	n = adfReadFile(file, EXTBUFL, extbuf);
	while(!adfEndOfFile(file)) {
		fwrite(extbuf, sizeof(unsigned char), n, out);
		n = adfReadFile(file, EXTBUFL, extbuf);
	}
	if (n>0)
		fwrite(extbuf, sizeof(unsigned char), n, out);

	fclose(out);

	adfCloseFile(file);
}

int main(int argc, char **argv)
{
	char strbuf[80];
	struct Device *dev;
	struct Volume *vol;
	unsigned char *extbuf;
	unsigned char adfName[256];
#ifdef PACKER
	int Type, Err = 0;
#endif

	/* get program name */
	if((progname = strrchr(argv[0], '/'))) progname++; else progname = argv[0];

#ifdef DEBUG
	openlog(progname, LOG_PID, LOG_USER);
	log("%s: %s", argv[2], argv[1]);
#endif
	/* check for parameters */
	if(argc == 1) {
		usage();
	}

	extbuf =(unsigned char*)malloc(EXTBUFL*sizeof(char));
	if (!extbuf) { fprintf(stderr,"malloc error\n"); exit(1); }

	strcpy(adfName, argv[2]);

#ifdef PACKER
	Type = VolGetFormat(adfName, &Err);
	if(Err) {
		sprintf(strbuf, "Can't unpack file '%s'.\n", argv[2]);
		fprintf(stderr, strbuf);
		exit(1);
	}
#endif
	/* initialize the library */
	adfEnvInitDefault();
	dev = adfMountDev( adfName,TRUE );
	if (!dev) {
		sprintf(strbuf,"Can't mount the dump device '%s'.\n", argv[2]);
		fprintf(stderr, strbuf);
		adfEnvCleanUp(); exit(1);
	}

	/* FIXME: How to pass volume number from MC ?? */
	vol = adfMount(dev, 0, TRUE);
	if (!vol) {
		fprintf(stderr, "Can't mount the volume\n");
		adfUnMountDev(dev);
		adfEnvCleanUp(); exit(1);
	}

	if(strcmp(argv[1], "view")==0 || strcmp(argv[1], "list") == 0) {
		/*
		 * show file list
		 */
		adffs_list(vol);
	} else if(strcmp(argv[1], "rm") == 0) {
		/*
		 * remove file
		 */
	} else if(strcmp(argv[1], "rmdir") == 0) {
		/*
		 * Remove directory
		 */
	} else if(strcmp(argv[1], "mkdir") == 0) {
		/*
		 * Create directory
		 */
	} else if(strcmp(argv[1], "copyin") == 0) {
		/*
		 * Copy file into adf file!
		 */
	} else if(strcmp(argv[1], "copyout") == 0) {
		/*
		 * Copy file out from adf file (aka: extract!)
		 */
		adffs_copyout(vol, argv[3], argv[4], extbuf);
	} else if(strcmp(argv[1], "run") == 0) {
		/*
		 * run file from archive !?
		 */
	} else {
		sprintf(strbuf, "Command %s is not supported!\n", argv[1]);
		fprintf(stderr, strbuf);
		adfUnMountDev(dev);
		adfEnvCleanUp(); exit(1);
	}

	adfUnMount(vol);
	adfUnMountDev(dev);

	adfEnvCleanUp();

	free(extbuf);

#ifdef DEBUG
	closelog();
#endif
	return 0;
}
